@@ -2,7 +2,6 @@ import os
 import os.path
 import subprocess
 import mimetypes
-import docassemble.base.filter
 import tempfile
 import shutil
 import fdfgen
@@ -11,24 +10,32 @@ import re
 import PyPDF2 as pypdf
 import pypdftk
 import string
+import codecs
 from PIL import Image
 from docassemble.base.error import DAError
 from docassemble.base.pdfa import pdf_to_pdfa
-from subprocess import call, check_output
 from docassemble.base.logger import logmessage
 from docassemble.base.functions import word
 from pdfminer.pdfparser import PDFParser
 from pdfminer.pdfdocument import PDFDocument
-from pdfminer.pdftypes import resolve1
+from pdfminer.pdftypes import resolve1, PDFObjRef
 from pdfminer.pdfpage import PDFPage
+import logging
+logging.getLogger('pdfminer').setLevel(logging.ERROR)
 import uuid
+from docassemble.base.config import daconfig
 
 PDFTK_PATH = 'pdftk'
+QPDF_PATH = 'qpdf'
 
 def set_pdftk_path(path):
     global PDFTK_PATH
     PDFTK_PATH = path
 
+def set_qpdf_path(path):
+    global QPDF_PATH
+    QPDF_PATH = path
+
 def read_fields(pdffile):
     import string
     printable = set(string.printable)
@@ -56,25 +63,35 @@ def fieldsorter(x):
         y_coord = 0
     return (x[2], y_coord, x_coord)
 
-def recursively_add_fields(fields, id_to_page, outfields):
+def recursively_add_fields(fields, id_to_page, outfields, prefix=''):
+    if isinstance(fields, PDFObjRef):
+        fields = resolve1(fields)
     for i in fields:
         field = resolve1(i)
-        name, value, rect, page, field_type = field.get('T'), field.get('V'), field.get('Rect'), field.get('P'), field.get('FT')
+        try:
+            name, value, rect, page, field_type = field.get('T'), field.get('V'), field.get('Rect'), field.get('P'), field.get('FT')
+        except:
+            logmessage("Skipping field " + repr(field))
+            continue
         if name is not None:
-            name = str(name).decode('latin1')
+            if not isinstance(name, bytes):
+                name = bytes(str(name), encoding='utf-8')
+            name = remove_nonprintable_bytes_limited(name)
         if value is not None:
-            value = str(value).decode('latin1')
-        #logmessage("name is " + str(name) + " and FT is |" + str(field_type) + "| and value is " + str(value))
+            if not isinstance(value, bytes):
+                value = bytes(str(value), encoding='utf-8')
+            value = remove_nonprintable_bytes_limited(value)
+        #logmessage("name is " + repr(name) + " and FT is |" + repr(str(field_type)) + "| and value is " + repr(value))
         if page is not None:
             pageno = id_to_page[page.objid]
         else:
             pageno = 1
-        if str(field_type) == '/Btn':
+        if str(field_type) in ('/Btn', "/'Btn'"):
             if value == '/Yes':
                 default = "Yes"
             else:
                 default = "No"
-        elif str(field_type) == '/Sig':
+        elif str(field_type) in ('/Sig', "/'Sig'"):
             default = '${ user.signature }'
         else:
             if value is not None:
@@ -89,16 +106,27 @@ def recursively_add_fields(fields, id_to_page, outfields):
                 default = word("something")
         kids = field.get('Kids')
         if kids:
-            recursively_add_fields(kids, id_to_page, outfields)
+            if name is None:
+                recursively_add_fields(kids, id_to_page, outfields, prefix=prefix)
+            else:
+                if prefix == '':
+                    recursively_add_fields(kids, id_to_page, outfields, prefix=name)
+                else:
+                    recursively_add_fields(kids, id_to_page, outfields, prefix=prefix + '.' + name)
         else:
-            outfields.append((name, default, pageno, rect, field_type))
+            if prefix != '' and name is not None:
+                outfields.append((prefix + '.' + name, default, pageno, rect, field_type))
+            elif prefix == '':
+                outfields.append((name, default, pageno, rect, field_type))
+            else:
+                outfields.append((prefix, default, pageno, rect, field_type))
 
 def read_fields_pdftk(pdffile):
-    output = unicode(check_output([PDFTK_PATH, pdffile, 'dump_data_fields']).decode('utf8'))
+    output = subprocess.check_output([PDFTK_PATH, pdffile, 'dump_data_fields']).decode()
     fields = list()
     if not len(output) > 0:
         return None
-    for field in yaml.load_all(output):
+    for field in yaml.load_all(output, Loader=yaml.FullLoader):
         if 'FieldType' in field and field['FieldType'] == 'Button':
             default = "No"
         else:
@@ -161,7 +189,7 @@ def recursive_add_bookmark(reader, writer, outlines, parent=None):
                 cur_bm = writer.addBookmark(destination.title, destination_page, parent, None, False, False, destination.typ)
             #logmessage("Added bookmark " + destination.title)
 
-def fill_template(template, data_strings=[], data_names=[], hidden=[], readonly=[], images=[], pdf_url=None, editable=True, pdfa=False, password=None):
+def fill_template(template, data_strings=[], data_names=[], hidden=[], readonly=[], images=[], pdf_url=None, editable=True, pdfa=False, password=None, template_password=None):
     if pdf_url is None:
         pdf_url = ''
     fdf = fdfgen.forge_fdf(pdf_url, data_strings, data_names, hidden, readonly)
@@ -177,20 +205,36 @@ def fill_template(template, data_strings=[], data_names=[], hidden=[], readonly=
 ".xfdf", delete=False)
         shutil.copyfile(xfdf_temp_filename, xfdf_file.name)
     pdf_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".pdf", delete=False)
+    if template_password is not None:
+        template_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".pdf", delete=False)
+        qpdf_subprocess_arguments = [QPDF_PATH, '--decrypt', '--password=' + template_password, template, template_file.name]
+        try:
+            result = subprocess.run(qpdf_subprocess_arguments, timeout=60).returncode
+        except subprocess.TimeoutExpired:
+            result = 1
+            logmessage("fill_template: call to qpdf took too long")
+        if result != 0:
+            logmessage("Failed to decrypt PDF template " + str(template))
+            raise DAError("Call to qpdf failed for template " + str(template) + " where arguments were " + " ".join(qpdf_subprocess_arguments))
+        template = template_file.name
     subprocess_arguments = [PDFTK_PATH, template, 'fill_form', fdf_file.name, 'output', pdf_file.name]
     #logmessage("Arguments are " + str(subprocess_arguments))
-    if editable:
+    if editable or len(images):
         subprocess_arguments.append('need_appearances')
     else:
         subprocess_arguments.append('flatten')
-    result = call(subprocess_arguments)
+    try:
+        result = subprocess.run(subprocess_arguments, timeout=600).returncode
+    except subprocess.TimeoutExpired:
+        result = 1
+        logmessage("fill_template: call to pdftk fill_form took too long")
     if result != 0:
         logmessage("Failed to fill PDF form " + str(template))
         raise DAError("Call to pdftk failed for template " + str(template) + " where arguments were " + " ".join(subprocess_arguments))
     if len(images):
         fields = dict()
         for field, default, pageno, rect, field_type in read_fields(template):
-            if str(field_type) == '/Sig':
+            if str(field_type) in ('/Sig', "/'Sig'"):
                 fields[field] = {'pageno': pageno, 'rect': rect}
         image_todo = list()
         for field, file_info in images:
@@ -199,8 +243,12 @@ def fill_template(template, data_strings=[], data_names=[], hidden=[], readonly=
                 continue
             #logmessage("Need to put image on page " + str(fields[field]['pageno']))
             temp_png = tempfile.NamedTemporaryFile(mode="wb", suffix=".png")
-            args = ["convert", file_info['fullpath'], "-trim", "+repage", temp_png.name]
-            result = call(args)
+            args = [daconfig.get('imagemagick', 'convert'), file_info['fullpath'], "-trim", "+repage", "+profile", '*', '-density', '0', temp_png.name]
+            try:
+                result = subprocess.run(args, timeout=60).returncode
+            except subprocess.TimeoutExpired:
+                logmessage("fill_template: convert took too long")
+                result = 1
             if result == 1:
                 logmessage("failed to trim file: " + " ".join(args))
                 continue
@@ -214,9 +262,13 @@ def fill_template(template, data_strings=[], data_names=[], hidden=[], readonly=
             else:
                 dpp = dppy
             extent_x, extent_y = xone*dpp+width, yone*dpp+height
-            overlay_pdf_file = tempfile.NamedTemporaryFile(mode="wb", suffix=".pdf", delete=False)
-            args = ["convert", temp_png.name, "-background", "none", "-density", str(int(dpp*72)), "-gravity", "NorthEast", "-extent", str(int(extent_x)) + 'x' + str(int(extent_y)), overlay_pdf_file.name]
-            result = call(args)
+            overlay_pdf_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".pdf", delete=False)
+            args = [daconfig.get('imagemagick', 'convert'), temp_png.name, "-background", "none", "-density", str(int(dpp*72)), "-gravity", "NorthEast", "-extent", str(int(extent_x)) + 'x' + str(int(extent_y)), overlay_pdf_file.name]
+            try:
+                result = subprocess.run(args, timeout=60).returncode
+            except subprocess.TimeoutExpired:
+                result = 1
+                logmessage("fill_template: call to convert took too long")
             if result == 1:
                 logmessage("failed to make overlay: " + " ".join(args))
                 continue
@@ -242,7 +294,7 @@ def fill_template(template, data_strings=[], data_names=[], hidden=[], readonly=
                 for i in range(original.getNumPages()):
                     newpage = original.getPage(i)
                     writer.addPage(newpage)
-                for key, val in tree.iteritems():
+                for key, val in tree.items():
                     writer._root_object.update({pypdf.generic.NameObject(key): val})
                 writer.page_list = list()
                 recursive_get_pages(writer._root_object['/Pages'], writer.page_list)
@@ -252,6 +304,8 @@ def fill_template(template, data_strings=[], data_names=[], hidden=[], readonly=
             shutil.copyfile(new_pdf_file.name, pdf_file.name)
             for item in image_todo:
                 item['overlay_stream'].close()
+    if (not editable) and len(images):
+        flatten_pdf(pdf_file.name)
     if pdfa:
         pdf_to_pdfa(pdf_file.name)
     if editable:
@@ -260,60 +314,18 @@ def fill_template(template, data_strings=[], data_names=[], hidden=[], readonly=
         pdf_encrypt(pdf_file.name, password)
     return pdf_file.name
 
-def concatenate_files(path_list, pdfa=False, password=None):
-    pdf_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".pdf", delete=False)
-    subprocess_arguments = [PDFTK_PATH]
-    new_path_list = list()
-    for path in path_list:
-        mimetype, encoding = mimetypes.guess_type(path)
-        if mimetype.startswith('image'):
-            new_pdf_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".pdf", delete=False)
-            args = ["convert", path, new_pdf_file.name]
-            result = call(args)
-            if result != 0:
-                logmessage("failed to convert image to PDF: " + " ".join(args))
-                continue
-            new_path_list.append(new_pdf_file.name)
-        elif mimetype in ('application/rtf', 'application/vnd.openxmlformats-officedocument.wordprocessingml.document', 'application/msword', 'application/vnd.oasis.opendocument.text'):
-            new_pdf_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".pdf", delete=False)
-            if mimetype == 'application/rtf':
-                ext = 'rtf'
-            elif mimetype == 'application/vnd.openxmlformats-officedocument.wordprocessingml.document':
-                ext = 'docx'
-            elif mimetype == 'application/msword':
-                ext = 'doc'
-            elif mimetype == 'application/vnd.oasis.opendocument.text':
-                ext = 'odt'
-            docassemble.base.pandoc.word_to_pdf(path, ext, new_pdf_file.name, pdfa=False)
-            new_path_list.append(new_pdf_file.name)
-        elif mimetype == 'application/pdf':
-            new_path_list.append(path)
-    if len(new_path_list) == 0:
-        raise DAError("concatenate_files: no valid files to concatenate")
-    subprocess_arguments.extend(new_path_list)
-    subprocess_arguments.extend(['cat', 'output', pdf_file.name])
-    #logmessage("Arguments are " + str(subprocess_arguments))
-    result = call(subprocess_arguments)
-    if result != 0:
-        logmessage("Failed to concatenate PDF files")
-        raise DAError("Call to pdftk failed for concatenation where arguments were " + " ".join(subprocess_arguments))
-    if pdfa:
-        pdf_to_pdfa(pdf_file.name)
-    replicate_js_and_calculations(new_path_list[0], pdf_file.name, password)
-    return pdf_file.name
-
 def get_passwords(password):
     if password is None:
         return (None, None)
     if type(password) in (str, unicode, bool, int, float):
-        owner_password = unicode(password).strip()
-        user_password = unicode(password).strip()
+        owner_password = str(password).strip()
+        user_password = str(password).strip()
     elif type(password) is list:
-        owner_password = unicode(password[0]).strip()
-        user_password = unicode(password[1]).strip()
+        owner_password = str(password[0]).strip()
+        user_password = str(password[1]).strip()
     elif type(password) is dict:
-        owner_password = unicode(password.get('owner', 'password')).strip()
-        user_password = unicode(password.get('user', 'password')).strip()
+        owner_password = str(password.get('owner', 'password')).strip()
+        user_password = str(password.get('user', 'password')).strip()
     else:
         raise DAError("get_passwords: invalid password")
     return (owner_password, user_password)
@@ -321,13 +333,13 @@ def get_passwords(password):
 def pdf_encrypt(filename, password):
     #logmessage("pdf_encrypt: running; password is " + repr(password))
     (owner_password, user_password) = get_passwords(password)
-    outfile = tempfile.NamedTemporaryFile(suffix=".pdf", delete=False)
+    outfile = tempfile.NamedTemporaryFile(prefix="datemp", suffix=".pdf", delete=False)
     if owner_password == user_password:
         commands = ['pdftk', filename, 'output', outfile.name, 'user_pw', user_password, 'allow', 'printing']
     else:
         commands = ['pdftk', filename, 'output', outfile.name, 'owner_pw', owner_password, 'user_pw', user_password, 'allow', 'printing']
     try:
-        output = subprocess.check_output(commands, stderr=subprocess.STDOUT)
+        output = subprocess.check_output(commands, stderr=subprocess.STDOUT).decode()
     except subprocess.CalledProcessError as err:
         output = err.output
         raise DAError("pdf_encrypt: error running pdftk.  " + output)
@@ -392,7 +404,7 @@ class DAPdfFileWriter(pypdf.PdfFileWriter):
         action = pypdf.generic.DictionaryObject()
         zoomArgs = []
         for a in args:
-            if a is not None:
+            if a is not None and a.__class__.__name__ != 'NullObject':
                 zoomArgs.append(pypdf.generic.NumberObject(a))
             else:
                 zoomArgs.append(pypdf.generic.NullObject())
@@ -435,12 +447,33 @@ class DAPdfFileWriter(pypdf.PdfFileWriter):
         return bookmarkRef
 
 def remove_nonprintable(text):
-    final = ''
+    final = str()
     for char in text:
         if char in string.printable:
             final += char
     return final
 
+def remove_nonprintable_bytes(byte_list):
+    final = str()
+    for the_int in byte_list:
+        if chr(the_int) in string.printable:
+            final += chr(the_int)
+    return bytearray(final, 'utf-8')
+
+def remove_nonprintable_bytes_limited(byte_list):
+    final = bytes()
+    if len(byte_list) >= 2 and byte_list[0] == 254 and byte_list[1] == 255:
+        byte_list = byte_list[2:]
+    for the_int in byte_list:
+        if the_int > 0:
+            final += bytes([the_int])
+    return codecs.decode(final, 'latin1')
+
+def remove_nonprintable_limited(text):
+    text = re.sub(r'^\xfe\xff', '', text)
+    text = re.sub(r'\x00', '', text)
+    return codecs.decode(text, 'latin1')
+
 def replicate_js_and_calculations(template_filename, original_filename, password):
     #logmessage("replicate_js_and_calculations where template_filename is " + template_filename + " and original_filename is " + original_filename + " and password is " + repr(password))
     template = pypdf.PdfFileReader(open(template_filename, 'rb'))
@@ -474,9 +507,9 @@ def replicate_js_and_calculations(template_filename, original_filename, password
                 js_obj = obj.getObject()
                 if '/S' in js_obj and js_obj['/S'] == '/JavaScript' and '/JS' in js_obj:
                     if isinstance(js_obj['/JS'], pypdf.generic.ByteStringObject) or isinstance(js_obj['/JS'], pypdf.generic.TextStringObject):
-                        js_to_write.append((name, remove_nonprintable(js_obj['/JS'])))
+                        js_to_write.append((name, remove_nonprintable_bytes(js_obj['/JS'])))
                     elif isinstance(js_obj['/JS'], pypdf.generic.EncodedStreamObject) or isinstance(js_obj['/JS'], pypdf.generic.DecodedStreamObject):
-                        js_to_write.append((name, remove_nonprintable(js_obj['/JS'].getData())))
+                        js_to_write.append((name, remove_nonprintable_bytes(js_obj['/JS'].getData())))
 
     if len(js_to_write) == 0 and len(co_field_names) == 0:
         #logmessage("Nothing to do here")
@@ -535,6 +568,53 @@ def replicate_js_and_calculations(template_filename, original_filename, password
             #logmessage("Passwords for encryption are " + str(user_password) + " and " + str(owner_password))
             writer.encrypt(str(user_password), owner_pwd=str(owner_password))
     outfile = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".pdf", delete=False)
-    writer.write(outfile)
-    outfile.flush()
+    with open(outfile.name, 'wb') as fp:
+        writer.write(fp)
     shutil.move(outfile.name, original_filename)
+
+def flatten_pdf(filename):
+    #logmessage("flatten_pdf: running")
+    outfile = tempfile.NamedTemporaryFile(prefix="datemp", suffix=".pdf", delete=False)
+    subprocess_arguments = [PDFTK_PATH, filename, 'output', outfile.name, 'flatten']
+    #logmessage("Arguments are " + str(subprocess_arguments))
+    try:
+        result = subprocess.run(subprocess_arguments, timeout=60).returncode
+    except subprocess.TimeoutExpired:
+        result = 1
+        logmessage("flatten_pdf: call to pdftk took too long")
+    if result != 0:
+        logmessage("Failed to flatten PDF form " + str(template))
+        raise DAError("Call to pdftk failed for template " + str(template) + " where arguments were " + " ".join(subprocess_arguments))
+    commands = []
+    shutil.move(outfile.name, filename)
+
+def overlay_pdf(main_file, logo_file, out_file, first_page=None, last_page=None, logo_page=None, only=None):
+    main_pdf = pypdf.PdfFileReader(main_file)
+    logo_pdf = pypdf.PdfFileReader(logo_file)
+    output_pdf = pypdf.PdfFileWriter()
+    if first_page is None or first_page < 1:
+        first_page = 1
+    if last_page is None or last_page < 1:
+        last_page = main_pdf.getNumPages()
+    if first_page > main_pdf.getNumPages():
+        first_page = main_pdf.getNumPages()
+    if last_page < first_page:
+        last_page = first_page
+    if logo_page is None or logo_page < 1:
+        logo_page = 1
+    if logo_page > logo_pdf.getNumPages():
+        logo_page = logo_pdf.getNumPages()
+    for page_no in range(first_page - 1, last_page):
+        if only == 'even':
+            if page_no % 2 == 0:
+                continue
+        elif only == 'odd':
+            if page_no % 2 != 0:
+                continue
+        page = main_pdf.getPage(page_no)
+        page.mergePage(logo_pdf.getPage(logo_page - 1))
+    for page_no in range(main_pdf.getNumPages()):
+        page = main_pdf.getPage(page_no)
+        output_pdf.addPage(page)
+    with open(out_file, 'wb') as fp:
+        output_pdf.write(fp)
